from argparse import ArgumentParser, Namespace
import logging
import mimetypes
import os.path
from ssl import CERT_NONE, SSLContext
import sys

from .api import API


def ls(args: Namespace, api: API) -> None:
    if not args.item:
        for a in api.albums():
            print(a.name)

        return

    for a in api.albums():
        if a.name != args.item:
            continue

        for p in api.album_photos(a):
            print(f'{p.id}\t{p.name}')

        return

    raise Exception(f'Unable to find album {args.item}')


def rm(args: Namespace, api: API) -> None:
    for a in api.albums():
        if args.album not in [a.name, a.id]:
            continue

        for p in api.album_photos(a):
            if args.photo not in [p.name, p.id]:
                continue

            api.album_photos_delete(a, [p])
            return

        raise Exception(f'Unable to find photo {args.photo}')

    raise Exception(f'Unable to find album {args.item}')


def upload(args: Namespace, api: API) -> None:
    # TODO: Use API.album(); extend to support lookup by ID
    album = None
    for a in api.albums():
        if args.album in [a.name, a.id]:
            album = a
            break

    assert album

    mime_type = mimetypes.guess_type(args.path)[0]
    assert mime_type

    with open(args.path, 'rb') as f:
        api.album_photo_upload(
            album,
            f,
            name=os.path.basename(args.path),
            mime_type=mime_type)


def main():
    ap = ArgumentParser()
    ap.add_argument(
        '-k', dest='validate_https', action='store_false', default=True,
        help='disable HTTPS certificate checking')
    # TODO: Get from Keychain
    ap.add_argument('-p', dest='password', help='Pix-Star password')
    ap.add_argument(
        '-u', dest='username', help='Pix-Star username, without @mypixstar.com')
    ap.add_argument(
        '-v', dest='verbosity', action='count', default=0,
        help='increase logging verbosity; can be used multiple times')

    sp = ap.add_subparsers(dest='subcommand')

    sp.add_parser('help', help='show help')

    ls_ap = sp.add_parser('ls', help='list things')
    ls_ap.add_argument(
        'item', nargs='?',
        help='album whose photos to list; if absent list albums')

    rm_ap = sp.add_parser('rm', help='remove things')
    rm_ap.add_argument('album', help='the album whose photo to remove')
    rm_ap.add_argument('photo', help='the photo to remove')

    upload_ap = sp.add_parser('upload', help='upload a file')
    upload_ap.add_argument('album', help='album to upload the photo to')
    upload_ap.add_argument('path', help='path to the file to upload')

    args = ap.parse_args()

    if args.subcommand in ['help', None]:
        ap.print_help()
        sys.exit(0)

    logging.basicConfig(
        style='{', format='{message}', stream=sys.stderr,
        level=logging.ERROR - args.verbosity * 10)

    ctx = None
    if not args.validate_https:
        ctx = SSLContext()
        ctx.verify_mode = CERT_NONE

    if not args.username:
        sys.stderr.write('Username: ')
        args.username = input().strip()

    if not args.password:
        sys.stderr.write('Password: ')
        args.password = input().strip()

    api = API(ssl_context=ctx)
    api.login(args.username, args.password)

    if args.subcommand == 'ls':
        ls(args, api)
    elif args.subcommand == 'rm':
        rm(args, api)
    elif args.subcommand == 'upload':
        upload(args, api)
    else:
        raise Exception(f'command {args.subcommand} not found')



from collections import namedtuple
from http.cookiejar import CookieJar
from ssl import SSLContext
from typing import IO, List, Optional
import lxml.etree
import random
import re
from string import ascii_lowercase
import time
from urllib.parse import urlencode
import urllib.request


Album = namedtuple('Album', ['id', 'name'])
Photo = namedtuple('Photo', ['id', 'name'])


class API:
    cookie_jar: CookieJar
    csrf_token: Optional[str]
    url_opener: urllib.request.OpenerDirector

    def __init__(self, ssl_context: SSLContext=None):
        self.cookie_jar = CookieJar()
        self.url_opener = urllib.request.build_opener(
            urllib.request.HTTPCookieProcessor(self.cookie_jar),
            urllib.request.HTTPSHandler(context=ssl_context))

    def login(self, username: str, password: str) -> None:
        '''
        Login to the Pix-Star service.

        This method must be invoked before any other methods.
        '''

        assert not self.csrf_token

        self.url_opener.open('https://www.pix-star.com/')
        for c in self.cookie_jar:
            if c.name == 'csrftoken':
                self.csrf_token = c.value

        assert self.csrf_token

        resp = self.url_opener.open(
            'https://www.pix-star.com/accounts/login/',
            data=urlencode({
                'csrfmiddlewaretoken': self.csrf_token,
                'username': username,
                'password': password}).encode('utf-8'))

        assert resp.status == 200
        assert resp.geturl() == 'https://www.pix-star.com/my_pixstar/'

    # TODO: Handle paging
    def albums(self) -> List[Album]:
        '''
        List the user's web albums.
        '''

        assert self.csrf_token

        resp = self.url_opener.open('https://www.pix-star.com/albums/list/')
        assert resp.status == 200

        return _parse_list_response(resp)

    def album(self, name: str) -> Album:
        '''
        Get an album with a specific name. Raises a KeyError if the album could
        not be found.
        '''

        assert self.csrf_token

        for a in self.albums():
            if a.name == name:
                return a

        raise KeyError()

    # TODO: How does the server handle duplicate album names? The UI pops up an alert
    #       but doesn't actually make the request
    def album_create(self, name: str) -> Album:
        '''
        Create the given album, returning it.
        '''

        qs = urlencode({
            'album_name': name,
            'csrfmiddlewaretoken': self.csrf_token,
        })
        resp = self.url_opener.open(
            f'https://www.pix-star.com/create/new_album/?{qs}')

        assert resp.status == 200
        m = re.match(r'^https://www.pix-star.com/album/web/([0-9]+)/', resp.geturl())
        assert m
        album_id = m.group(1)

        for a in self.albums():
            if a.id == album_id:
                return a

        raise Exception('Could not find newly-created album')

    def album_delete(self, albums: List[Album]) -> None:
        '''
        Delete several albums.
        '''

        # TODO: Paging of delete requests
        req = urllib.request.Request(
            f'https://www.pix-star.com/delete/album/',
            data=urlencode(
                [('web_albums[]', a.id) for a in albums]).encode('utf-8'))
        req.add_header('X-Requested-With', 'XMLHttpRequest')

        resp = self.url_opener.open(req)
        assert resp.status == 200

    def album_photos(self, album: Album) -> List[Photo]:
        '''
        Get information about the given album.
        '''

        assert self.csrf_token

        photos: List[Photo] = []
        page_num = 1
        while True:
            qs = urlencode({
                'page': page_num,
                'size': 'small',
                '_': int(time.time() * 1000),
            })
            req = urllib.request.Request(f'https://www.pix-star.com/album/web/{album.id}/?{qs}')
            # XXX: Without this X-Requested-With header, we get a 404 on the last page
            #      rather than the sentintel 'no-more' response. The Pix-Star website
            #      uses the sentinel response, so do the same here.
            req.add_header('X-Requested-With', 'XMLHttpRequest')
            resp = self.url_opener.open(req)
            assert resp.status == 200

            page_photos = _parse_album_photos_response(resp)
            if not page_photos:
                return photos

            photos += page_photos
            page_num += 1

    def album_photo_upload(self, album: Album, f: IO, name: str, mime_type: str) -> None:
        '''
        Upload a photo to an album.
        '''

        mime_boundary = '----pyxstar'
        mime_boundary += ''.join(random.choices(ascii_lowercase, k=16))

        body = bytearray()
        body += '\r\n'.join([
            ('--' + mime_boundary),
            f'Content-Disposition: form-data; name="image"; filename="{name}"',
            f'Content-Type: {mime_type}',
            '',
            '']).encode('utf-8')
        body += f.read() + b'\r\n'
        body += ('--' + mime_boundary + '--').encode('utf-8') + b'\r\n'

        qs = urlencode({'size': 'small'})
        req = urllib.request.Request(
            f'https://www.pix-star.com/album/web/{album.id}/?{qs}',
            data=body)
        req.add_header('Content-Type', 'multipart/form-data; boundary=' + mime_boundary)
        req.add_header('X-Requested-With', 'XMLHttpRequest')

        resp = self.url_opener.open(req)
        assert resp.status == 200

    def album_photos_delete(self, album: Album, photos: List[Photo]):
        '''
        Delete photos from the given album.
        '''

        # TODO: Paging of delete requests
        req = urllib.request.Request(
            f'https://www.pix-star.com/delete/album/image/',
            data=urlencode(
                [('images[]', p.id) for p in photos] +
                [
                    ('album_id', album.id),
                    ('size', 'small'),
                    ('album_type', 'web')]).encode('utf-8'))
        req.add_header('X-Requested-With', 'XMLHttpRequest')

        resp = self.url_opener.open(req)
        assert resp.status == 200


def _parse_list_response(f: IO) -> List[Album]:
    albums = []

    doc = lxml.etree.parse(f, lxml.etree.HTMLParser())

    # TODO: Use CSS selectors which will handle attributes better, e.g. not requiring
    #       a full class match. See https://lxml.de/apidoc/lxml.cssselect.html.
    for a in doc.xpath('//a[@class="album_title"]'):
        parent = a.getparent()
        assert parent.tag == 'div'

        id = parent.attrib['id']
        assert id.startswith('album_id_')
        id = id[len('album_id_'):]

        albums.append(Album(id=id, name=a.text))

    return albums


def _parse_album_photos_response(f: IO) -> List[Photo]:
    photos: List[Photo] = []

    # There are some sentinel values that the "API" returns -- 'no-data' for a
    # read past the last page of results, and '\n\n' when the album is
    # completely empty.
    data = f.read().strip()
    if not data or data == 'no-more':
        return photos

    doc = lxml.etree.fromstring(data, lxml.etree.HTMLParser())

    for li in doc.xpath('//li[@class="album-image"]'):
        cb = li.xpath('.//input[@type="checkbox"][@class="image-checkbox"]')
        assert len(cb) == 1

        a = li.xpath('.//h5[@class="photo_title"]')
        assert len(a) == 1

        photos.append(
            Photo(
                id=cb[0].attrib['value'],
                name=a[0].attrib['title']))

    return photos



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyxstar']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.5.0,<5.0.0']

entry_points = \
{'console_scripts': ['pyxstar = pyxstar.main:main']}

setup_kwargs = {
    'name': 'pyxstar',
    'version': '0.2.2',
    'description': 'A library and commandline tool for managing Pix-Star digital photo frames',
    'long_description': "A Python library and commandline tool for managing [Pix-Star](https://www.pix-star.com/) digital photo frames.\n\n# Installation\n\nYou can install from PyPI using `pip` as follows\n\n```bash\npip install pyxstar\n```\n\n# Usage\n\n## Python library\n\nThe `pyxstar.api` module is the basis for all API interactions with the Pix-Star service. The `API` class should be used to invoke methods on the service, which accept and return `Album` and `Photo` classes.\n\nFor example\n\n```python\nfrom pyxstar.api import API\n\napi = API()\napi.login('myusername', 'mypassword')\nfor a in api.albums():\n    print(f'Album: {a.name}')\n\n    for p in api.album_photos(a):\n        print(f'  Photo: {p.name}')\n```\n\n## Commandline\n\nThis package provides a `pyxstar` commandline tool which offers a variety of subcommands to interact with your digital photo frame.\n\nThe following are some examples of how to use this:\n\n```bash\n# Show help\n$ pyxstar help\n[...]\n\n# List album names\n$ pyxstar -u myusername -p mypassword ls\nMy First Album\nMy Second Album\n\n# List photos in My First Album\n$ pyxstar -u myusername -p mypassword ls 'My First Album'\n315371094   _dsc1254_59.jpg\n315371095   _dsc1254_60.jpg\n\n# Upload a photo to My First Album and check that it exists\n$ pyxstar -u myusername -p mypassword upload 'My First Album' /path/to/foo.jpg\n$ pyxstar -u myusername -p mypassword ls 'My First Album'\n315371094   _dsc1254_59.jpg\n315371095   _dsc1254_60.jpg\n315371099   foo.jpg\n```\n",
    'author': 'Peter Griess',
    'author_email': 'pg@std.in',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pgriess/pyxstar',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
